/*-
 * Copyright (C)2022 @BABOLO http://www.babolo.ru/
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ident "@(#) Copyright (C)2022 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: mod_authn_sqlite.c,v 1.4 2022/05/15 00:06:17 babolo Exp $"

#include <apr_strings.h>
#include <apr_md5.h>
#include <ap_config.h>
#include <ap_provider.h>
#include <httpd.h>
#include <sqlite.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <mod_auth.h>

#define MAXPWSZ 100

typedef struct {
    char *dbfile;
    char *passwd;
    char *digest;
    char *context;
    char *allpass;
    char pop[];
} authn_sqlite_config_rec;

module AP_MODULE_DECLARE_DATA authn_sqlite_module;

const static char f[] =
    "SELECT cred\n"
    " FROM \"%s\"\n"
    " WHERE login = ? AND (passwd = ? OR passwd IS NULL)\n"
    "ORDER BY cred\n"
    "LIMIT 1\n"
    ";";

const static char c[] =
    "SELECT cred\n"
    " FROM \"%s\"\n"
    " WHERE login = ? AND (passwd = ? OR passwd IS NULL) AND context = ?\n"
    "ORDER BY cred\n"
    "LIMIT 1\n"
    ";";

const static char u[] =
    " *+,-./0123456789:=@ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";

static const command_rec authn_sqlite_cmds[] =
{   AP_INIT_TAKE1( "Auth_SQLite_File"
                 , ap_set_file_slot
                 , (void *)APR_OFFSETOF(authn_sqlite_config_rec, dbfile)
                 , OR_AUTHCFG
                 , "SQLite DB file containing user IDs and passwords"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite_Passwd"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite_config_rec, passwd)
                 , OR_AUTHCFG
                 , "Name of passwd table"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite_Digest"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite_config_rec, digest)
                 , OR_AUTHCFG
                 , "Name of digest table"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite_Context"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite_config_rec, context)
                 , OR_AUTHCFG
                 , "Context of AAA"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite_AllPass"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite_config_rec, allpass)
                 , OR_AUTHCFG
                 , "All pass URI begin"
                 )
,   {NULL}
};

static void
testset(request_rec *r) {
    if  (!!r->connection->client_ip)
        apr_table_setn(r->subprocess_env, "TEST_CIP"  , r->connection->client_ip  );
    if  (!!r->connection->remote_host)
        apr_table_setn(r->subprocess_env, "TEST_RHOST", r->connection->remote_host);
    if  (!!r->connection->local_ip)
        apr_table_setn(r->subprocess_env, "TEST_LIP"  , r->connection->local_ip   );
    if  (!!r->connection->local_host)
        apr_table_setn(r->subprocess_env, "TEST_LHOST", r->connection->local_host );
    if  (!!r->uri)            apr_table_setn(r->subprocess_env, "TEST_URI"  , r->uri           );
    if  (!!r->args)           apr_table_setn(r->subprocess_env, "TEST_ARGS" , r->args          );
    if  (!!r->protocol)       apr_table_setn(r->subprocess_env, "TEST_PROTO", r->protocol      );
    if  (!!r->the_request)    apr_table_setn(r->subprocess_env, "TEST_REQ"  , r->the_request   );
    if  (!!r->hostname)       apr_table_setn(r->subprocess_env, "TEST_HOST" , r->hostname      );
    if  (!!r->method)         apr_table_setn(r->subprocess_env, "TEST_METH" , r->method        );
    if  (!!r->user)           apr_table_setn(r->subprocess_env, "TEST_USER" , r->user          );
    if  (!!r->unparsed_uri)   apr_table_setn(r->subprocess_env, "TEST_UURI" , r->unparsed_uri  );
    if  (!!r->path_info)      apr_table_setn(r->subprocess_env, "TEST_PATH" , r->path_info     );
    if  (!!r->useragent_ip)   apr_table_setn(r->subprocess_env, "TEST_UIP"  , r->useragent_ip  );
    if  (!!r->useragent_host) apr_table_setn(r->subprocess_env, "TEST_UHOST", r->useragent_host);
}

static void *
create_authn_sqlite_dir_config(apr_pool_t *p, char *d) {
    authn_sqlite_config_rec *conf;

    conf = apr_palloc(p, sizeof(authn_sqlite_config_rec) + strlen(c) + MAXPWSZ);
    bzero(conf, sizeof(authn_sqlite_config_rec) + strlen(c));
    return(conf);
}

static authn_status
check_password(request_rec *r, const char *user, const char *password) {
    char                    *errmsg = NULL;
    const char             **tuple;
    authn_sqlite_config_rec *conf   = ap_get_module_config(r->per_dir_config, &authn_sqlite_module);
    int                      ncol;
    sqlite                  *odf;
    sqlite_vm               *vm;
    size_t                   sz = 0;
    int                      ex;
    int                      i;

    if  (!conf->dbfile) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Auth_SQLite_File not specified in the configuration"
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->passwd) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Auth_SQLite_Passwd not specified in the configuration"
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->pop[0]) {
        if  (!!conf->passwd[strspn(conf->passwd, u)]) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite@BABOLO Illegal table name: %s"
                         , conf->passwd
                         );
            return(AUTH_GENERAL_ERROR);
        }
        sz = strlen((!!conf->context) ? c : f) + strlen(conf->passwd);
        if  (sz > strlen(c) + MAXPWSZ) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite@BABOLO Too big tablename %s"
                         , conf->passwd
                         );
            return(AUTH_GENERAL_ERROR);
        }
        snprintf(conf->pop, sz + 1, (!!conf->context) ? c : f, conf->passwd);
    }
    if  (!(odf = sqlite_open(conf->dbfile, 0, NULL))) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Could not open password DB: %s"
                     , conf->dbfile
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite_compile(odf, conf->pop, NULL, &vm, NULL))) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Could not compile %d<%s>: %s"
                     , ex
                     , sqlite_error_string(ex)
                     , conf->pop
                     );
        if  (!vm || !!sqlite_finalize(vm, &errmsg)) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite@BABOLO Could not compile: %s"
                         , !!errmsg ? errmsg : ""
                         );
        }
        return(AUTH_GENERAL_ERROR);
    }
    i = 0;
    if  (!!conf->context && !!(ex = sqlite_bind(vm, ++i, conf->context, -1, 0))) {
        sqlite_close(odf);
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Could not bind[%d] %d<%s>: %s"
                     , i
                     , ex
                     , sqlite_error_string(ex)
                     , conf->context
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite_bind(vm, ++i, user, -1, 0))) {
        sqlite_close(odf);
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Could not bind[%d] %d<%s>: %s"
                     , i
                     , ex
                     , sqlite_error_string(ex)
                     , user
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite_bind(vm, ++i, password, -1, 0))) {
        sqlite_close(odf);
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Could not bind[%d] %d<%s>: %s"
                     , i
                     , ex
                     , sqlite_error_string(ex)
                     , password
                     );
        return(AUTH_GENERAL_ERROR);
    }
    switch((ex = sqlite_step(vm, &ncol, &tuple, NULL))) {
    case SQLITE_ROW:
        if  (ncol != 1) {
            sqlite_finalize(vm, NULL);
            sqlite_close(odf);
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite@BABOLO Col num %d IN %s"
                         , ncol
                         , conf->pop
                         );
            return(AUTH_GENERAL_ERROR);
        }
        if  (!!tuple) apr_table_setn(r->subprocess_env, "CREDENTIALS", *tuple);
        testset(r);
        sqlite_finalize(vm, NULL);
        sqlite_close(odf);
        return(AUTH_GRANTED);
    case SQLITE_DONE:
        sqlite_finalize(vm, NULL);
        sqlite_close(odf);
        if  (!!conf->allpass && !strcmp(r->uri, conf->allpass)) {
            apr_table_setn(r->subprocess_env, "CREDENTIALS", "0");
            testset(r);
            return(AUTH_GRANTED);
        }
        return(AUTH_DENIED);
    }
    sqlite_finalize(vm, NULL);
    sqlite_close(odf);
    ap_log_rerror( APLOG_MARK
                 , APLOG_ERR
                 , 0
                 , r
                 , "mod_authn_sqlite@BABOLO Could not SELECT %d<%s>"
                 , ex
                 , sqlite_error_string(ex)
                 );
    return(AUTH_GENERAL_ERROR);
}

static authn_status
get_realm_hash(request_rec *r, const char *user, const char *realm, char **rethash) {
    char                    *sqlite_hash = NULL;
    authn_sqlite_config_rec *conf = ap_get_module_config(r->per_dir_config, &authn_sqlite_module);

    if  (!conf->dbfile) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite@BABOLO Auth_SQLite_File not specified in the configuration"
                     );
        return AUTH_GENERAL_ERROR;
    }




    if (!sqlite_hash) return(AUTH_USER_NOT_FOUND);
    *rethash = sqlite_hash;
    return(AUTH_USER_FOUND);
}

static const authn_provider authn_sqlite_provider =
{   &check_password
,   &get_realm_hash
};

static void
register_hooks(apr_pool_t *p) {
    ap_register_auth_provider( p
                             , AUTHN_PROVIDER_GROUP
                             , "sqlite"
                             , AUTHN_PROVIDER_VERSION
                             , &authn_sqlite_provider
                             , AP_AUTH_INTERNAL_PER_CONF
                             );
}

AP_DECLARE_MODULE(authn_sqlite) =
{   STANDARD20_MODULE_STUFF
,   create_authn_sqlite_dir_config  /* dir config creater */
,   NULL                            /* dir merger --- default is to override */
,   NULL                            /* server config */
,   NULL                            /* merge server config */
,   authn_sqlite_cmds               /* command apr_table_t */
,   register_hooks                  /* register hooks */
};
